/*
	Simple TCP SOCKET INHERITANCE TEST
	Parent binary (This one) binary opens a TCP socket pair, then launches CHLD_PROC.EXE sending the listening socket handle value on 
	command line, It then pauses and sends data. Child uses handle inheritance CHILD ONLY CALLS WSASTARTUP-RECV-WSACLEANUP
*/

#include<stdio.h>
#include<unistd.h> /* for sleep() */
#include<io.h>
#include<corerror.h>
#include<winsock2.h>

#define BUFLEN 512	//Max length of buffer
#define PORT 8888	//The port on which to listen for incoming data

int main()
{
	SOCKET /* su, */ st, ct, xt; // server udp, server tcp, client tcp and connection accepted tcp sockets
	struct sockaddr_in server, si_other;
	int slen , recv_len;
	char buf[BUFLEN];
	char cmd[BUFLEN];
	WSADATA wsa;
	HANDLE parentHandle = OpenProcess(MAXIMUM_ALLOWED, FALSE, GetCurrentProcessId());
	if ( ! SetHandleInformation(parentHandle, HANDLE_FLAG_INHERIT, 1) )
      		printf("Failed to set Parent Process handle as inheritable!"); 
	slen = sizeof(si_other) ;
	
	//Initialise winsock
	printf("\nWinsock socket inheritance test starting up\nInitialising Winsock... ");
	if (WSAStartup(MAKEWORD(2,2),&wsa) != 0)
	{
		printf("Winsock Init Failed. Error Code : %d\n",WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("Winsock Initialized on Parent process.\n");

	//Create a TCP socket
	if((st = socket(AF_INET , SOCK_STREAM , 0 )) == INVALID_SOCKET)
	{
		printf("Could not create socket : %d" , WSAGetLastError());
	}
	printf("TCP server Socket created, ");
	
	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( PORT ); // same port, but we are on TCP this time so there is no conflict
	
	//Bind
	if( bind(st ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("Bind completed, ");

	// start TCP socket listening
	if( listen(st ,SOMAXCONN) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}	printf("now Listening on 8888 (visible in netstat).\n");

	// now we create yet another TCP socket but specifying port zero to get a dynamic port number
	
	//Create a TCP socket
	if((ct = socket(AF_INET , SOCK_STREAM , 0 )) == INVALID_SOCKET)
	{
		printf("Could not create TCP client socket error: %d\n" , WSAGetLastError());
	}
	printf("TCP client Socket created, ");
	
	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr("127.0.0.1"); // not sure this will work, may need to prompt user for host IP
	server.sin_port = htons( 0 ); // port zero to get dynamic port assigned.. don't care
	
	//Bind
	if( bind(ct ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("client Bind completed using dynamic port, ");

	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr("127.0.0.1"); // not sure this will work, may need to prompt user for host IP
	server.sin_port = htons( PORT ); // connect to our target socket which is already listening

	if( connect(ct ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Connect failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("client socket Connected.\n");

	// the client is non-blocking, so we can now accept its connection from the server socket to get our conencted socket..	 
	if( (xt = accept(st , (struct sockaddr *) &si_other, &slen)) == SOCKET_ERROR)
	{
		printf("TCP Accept failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("accept completed new socket: %d - ", xt);
	printf("Received self-connect from %s:%d\n", inet_ntoa(si_other.sin_addr), ntohs(si_other.sin_port));
	
	// a loopback connection should now exist, we need to test this to see if it works before spawning a child that inherits it.
	if( send(ct , "TESTING 1..2..3", 15, 0) == SOCKET_ERROR)
	{
		printf("TCP Send failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("client send call succesful to listening port -");
	// since the send socket is non blocking the data we sent should be sitting in an OS buffer an ready to read if everything worked.
	//clear the buffer by filling null, it might have previously received data
	memset(buf,'\0', BUFLEN);
	//wait to receive some data, this is a blocking call
	if ((recv_len = recvfrom(xt, buf, BUFLEN, 0, (struct sockaddr *) &si_other, &slen)) == SOCKET_ERROR)
	{
		printf("recvfrom() failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("recvfrom successful, loopback responding\n  Loopback test Data received: %s\n", buf);

	// now launch the child application with handle inheritance. Since it will share the same console we need 
	// to warn the user that they will see output from both client and server in this same window from here forward.

	printf("\nLaunching test peer as 'child' process sharing window/socket by inheriting handles.\n");
	printf("From this point forward all messages will be prefixed with PARENT: or CHILD: (message source).\n\n");

	// test file handle inheritance too
//    	const char* filename = "fhtest.txt";
//	const char *str = "Arbitrary string to be written to a file.\n";
//	int fd = 0;
//    	fd = open(filename, O_RDWR | O_CREAT | O_TRUNC );
//    	if (fd == -1) {
//        	perror("open");
//        	exit(EXIT_FAILURE);
//    	}
//
//    	write(fd, str, strlen(str));
//    	printf("Done Writing!\n");
    HANDLE hFile; 
    char DataBuffer[] = "This is some test data to write to the file.";
    DWORD dwBytesToWrite = (DWORD)strlen(DataBuffer);
    DWORD dwBytesWritten = 0;
    BOOL bErrorFlag = FALSE;

    hFile = CreateFile("fhtest.txt",                // name of the write
                       GENERIC_WRITE,          // open for writing
                       0,                      // do not share
                       NULL,                   // default security
                       OPEN_ALWAYS, // CREATE_NEW,             // create new file only
                       FILE_ATTRIBUTE_NORMAL,  // normal file
                       NULL);                  // no attr. template

    if (hFile == INVALID_HANDLE_VALUE) 
    { 
        printf(TEXT("PARENT: ERROR - Unable to open file \"%s\" for write.\n"), "fhtest.txt");
        //return;
    }

    if ( ! SetHandleInformation(hFile, HANDLE_FLAG_INHERIT, 1) )
    	printf("PARENT: Failed to set File handle as inheritable!"); 


    printf(TEXT("PARENT: Writing %ld bytes to %s.\n"), dwBytesToWrite, "fhtest.txt");

    bErrorFlag = WriteFile( 
                    hFile,           // open file handle
                    DataBuffer,      // start of data to write
                    dwBytesToWrite,  // number of bytes to write
                    &dwBytesWritten, // number of bytes that were written
                    NULL);            // no overlapped structure

    if (FALSE == bErrorFlag)
    {
        printf("PARENT: ERROR - Unable to write to file.\n");
    }
    else
    {
        if (dwBytesWritten != dwBytesToWrite)
        {
            // This is an error because a synchronous write that results in
            // success (WriteFile returns TRUE) should write all data as
            // requested. This would not necessarily be the case for
            // asynchronous writes.
            printf("PARENT: Error - dwBytesWritten != dwBytesToWrite\n");
        }
        else
        {
            printf("PARENT: Wrote %ld bytes to %s successfully.\n", dwBytesWritten, "fhtest.txt");
        }
    }

// THIS IS NOT NEEDED ON WIN 10 but we'll try it in ROS to see if it helps.

    	if ( ! SetHandleInformation( (HANDLE)xt, HANDLE_FLAG_INHERIT, 1) )
    		printf("PARENT: Failed to set Accept Socket handle as inheritable!"); 

	// there is not much to this, but we use a parameter that is typically left as zero/false/null to tell windows to
	// let the child use our sockets as he sees fit.
        STARTUPINFO            sinfo;        
        PROCESS_INFORMATION    pinfo;        
        memset(&sinfo, 0, sizeof(STARTUPINFO));
        memset(&pinfo, 0, sizeof(PROCESS_INFORMATION));
        sinfo.dwFlags = STARTF_USESHOWWINDOW;
        sinfo.wShowWindow = SW_SHOWMAXIMIZED;

	memset(cmd,'\0', BUFLEN);
	sprintf(cmd, "CHLD_PROC.exe %d %d %d" , xt, (int)hFile, (int)parentHandle); 
	// ADDED file desriptor and parentHandle to test inheritance with file and process handles
	printf("PARENT: Executing child command: %s\n", cmd);
	// fifth parameter being TRUE tells windows we are inheriting handles, sockets are
	// inheritable by default so no other chane i needed to make them available in child
  	if (CreateProcess(NULL, cmd, NULL, NULL, TRUE, 0, NULL, NULL, &sinfo, &pinfo)) { 
    		printf("PARENT: launched CHLD_PROC.exe as child process successfully.\n");
  	} else {
    		printf("PARENT: Failed to launch CHLD_PROC.exe as child process...\n");
	}

//	memset(cmd,'\0', BUFLEN);
//	sprintf(cmd, "SOCKET:%d" , xt);
	printf("PARENT: pausing for child setup before sending data.\n");
//	recv_len = strlen(cmd);
	// now pause and then send using our connected TCP client socket to see if the client gets our data or not.
	Sleep(50);
	printf("\nPARENT: Sending test value (SOCKYOU) via TCP to see if Client can receive on inherited socket.\n");
	if( send(ct , "SOCKYOU", 7, 0) == SOCKET_ERROR)
	{
		printf("PARENT: TCP Send failed with error code : %d\n" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("\nPARENT: TEST ACTIVITIES COMPLETED. Press <any key> to exit.\n");
	//Sleep(50);
	getchar(); // NOTE: getchar prevents premature handle closing.. it is not just an optional pause.
	printf("PARENT: teardown and exit\n");
	//closesocket(su);
	closesocket(ct);
	closesocket(xt); // this one may be closed by previous line.. not sure
	closesocket(st);
//    	close(fd);
    	CloseHandle(hFile);
	CloseHandle(parentHandle);
	WSACleanup();
	return 0;		
}
