/*
	Simple UDP Server
*/

#include<stdio.h>
#include<unistd.h> /* for sleep() */
#include<winsock2.h>

/* pragma comment(lib,"ws2_32.lib") //Winsock Library */

#define BUFLEN 512	//Max length of buffer
#define PORT 8888	//The port on which to listen for incoming data

int main()
{
	SOCKET su, st, ct, xt; // server udp, server tcp, client tcp and connection accepted tcp sockets
	struct sockaddr_in server, si_other;
	int slen , recv_len;
	char buf[BUFLEN];
	char cmd[BUFLEN];
	char * ourCmd;
	WSADATA wsa;
	char cmd1[] = "SOCKME";
	char cmd2[] = "EXIT";

	slen = sizeof(si_other) ;
	
	//Initialise winsock
	printf("\nUDP Datagram and socket inheritance test server starting up\nInitialising Winsock...");
	if (WSAStartup(MAKEWORD(2,2),&wsa) != 0)
	{
		printf("Failed. Error Code : %d",WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("Winsock Initialized on Server App.\n");
	
	//Create a socket
	if((su = socket(AF_INET , SOCK_DGRAM , 0 )) == INVALID_SOCKET)
	{
		printf("Could not create socket : %d" , WSAGetLastError());
	}
	printf("UDP server Socket created.\n");
	
	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( PORT );
	
	//Bind
	if( bind(su ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("UDP server Bind done");

	// now we open a TCP socketpair for inheritance testing, this will be a TCP listening(server) socket and a TCP client socket
	// the 'client socket will be connected to the listener, and perform a send to show it can receive.
	// then we will execute the client application as a child process with handle inheritance set as true.

	// once that setup is confirmed and client launched the UDP side will wait for the incoming message to be 'SOCKME'
	// and once it comes in, the server app (us) will respond on UDP with the text SOCKET:<TCP Listening (server) socket number>
	// then it will wait for 5 seconds and execute a blocking TCP send sending the string 'SOCKYOU'
	// if the client gets it then inheritance is working, otherwise it is not.

	//Create a TCP socket
	if((st = socket(AF_INET , SOCK_STREAM , 0 )) == INVALID_SOCKET)
	{
		printf("Could not create socket : %d" , WSAGetLastError());
	}
	printf("TCP server Socket created.\n");
	
	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = INADDR_ANY;
	server.sin_port = htons( PORT ); // same port, but we are on TCP this time so there is no conflict
	
	//Bind
	if( bind(st ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP socket Bind done");

	// start TCP socket listening
	if( listen(st ,SOMAXCONN) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP socket Listen done, the TCP port should now be visible in netstat");

	// now we create yet another TCP socket but specifying port zero to get a dynamic port number
	
	//Create a TCP socket
	if((ct = socket(AF_INET , SOCK_STREAM , 0 )) == INVALID_SOCKET)
	{
		printf("Could not create socket : %d" , WSAGetLastError());
	}
	printf("TCP client Socket created.\n");
	
	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr("127.0.0.1"); // not sure this will work, may need to prompt user for host IP
	server.sin_port = htons( 0 ); // port zero to get dynamic port assigned.. don't care
	
	//Bind
	if( bind(ct ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
		printf("Bind failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP client socket Bind completed using dynamic port ");


	//gets(buf);

	//Prepare the sockaddr_in structure
	server.sin_family = AF_INET;
	server.sin_addr.s_addr = inet_addr("127.0.0.1"); // not sure this will work, may need to prompt user for host IP
	server.sin_port = htons( PORT ); // connect to our target socket which is already listening

	if( connect(ct ,(struct sockaddr *)&server , sizeof(server)) == SOCKET_ERROR)
	{
      		//if ((Total = select(0, &ReadSet, &WriteSet, NULL, NULL)) == SOCKET_ERROR)
      		//{
         	//	printf("select() returned with error %d\n", WSAGetLastError());
		//}
		printf("Connect failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP client socket Connect successful using dynamic port ");

/*
	//set client socket as nonblocking
	u_long mode = 1;  // 1 to enable non-blocking socket
	if( ioctlsocket(ct, FIONBIO, &mode) == SOCKET_ERROR)
	{
		printf("Failed to set TCP client socket Non Blocking with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP client socket set Non Blocking, check netstat and press key to continue... ");
*/

	// the client is non-blocking, so we can now accept its connection from the server socket to get our conencted socket..	 
	if( (xt = accept(st , (struct sockaddr *) &si_other, &slen)) == SOCKET_ERROR)
	{
		printf("TCP Accept failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	printf("TCP client socket conneciton accepted on socket: %d\n", xt);
	printf("Received connection from %s:%d\n", inet_ntoa(si_other.sin_addr), ntohs(si_other.sin_port));
	
	// a loopback connection should now exist, we need to test this to see if it works before spawning a child that inherits it.
	if( send(ct , "TESTING 1..2..3", 15, 0) == SOCKET_ERROR)
	{
		printf("TCP Send failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP client socket succesfully used to send to TCP server port ");

	// since the send socket is non blocking the data we sent should be sitting in an OS buffer an ready to read if everything worked.
	//clear the buffer by filling null, it might have previously received data
	memset(buf,'\0', BUFLEN);
	//wait to receive some data, this is a blocking call
	if ((recv_len = recvfrom(xt, buf, BUFLEN, 0, (struct sockaddr *) &si_other, &slen)) == SOCKET_ERROR)
	{
		printf("recvfrom() failed with error code : %d" , WSAGetLastError());
		exit(EXIT_FAILURE);
	}
	puts("TCP server socket recvfrom returned data so the loopback pipe is responding ");
	printf("Loopback Data Received: %s\n", buf);

	// now exec the upd client application with handle inheritance. Since it will share the same console we need 
	// to warn the user that they will see output from both client and server in this same window from here forward.

	puts("Launching the client test application as a 'child' process sharing this window for output by iinheriting handles.");
	puts("From this point forward all messages will be prefixed with SERVER: or CLIENT: to identify their source.");
	fflush(stdout);

	// there is not much to this, but we use a parameter that is typically left as zero/false/null to tell windows to
	// let the child use our sockets as he sees fit.

        STARTUPINFO            sinfo;        
        PROCESS_INFORMATION    pinfo;        
        memset(&sinfo, 0, sizeof(STARTUPINFO));
        memset(&pinfo, 0, sizeof(PROCESS_INFORMATION));

        sinfo.dwFlags = STARTF_USESHOWWINDOW;
        sinfo.wShowWindow = SW_SHOWMAXIMIZED;

  	if (CreateProcess(NULL, "UDP_Client.exe", NULL, NULL, TRUE, 0, NULL, NULL, &sinfo, &pinfo)) { /*si, pi) ) { */
    		puts("SERVER: launched UDP_Client.exe as child process successfully.");
  	} else {
    		puts("SERVER: Failed to launch UDP_Client.exe as child process...");
	}

	// Other than the command 'SOCKME' to start that test the server will also watch for the command 'EXIT' and when it
	// arrives it echoes as always and then the server application exits. The client will do the same thing when it sees 'EXIT'.

	//start UDP server listening for data
	while(1)
	{
		printf("SERVER: Waiting for data...\n");
		fflush(stdout);
		
		//clear the buffer by filling null, it might have previously received data
		memset(buf,'\0', BUFLEN);
		
		//try to receive some data, this is a blocking call
		if ((recv_len = recvfrom(su, buf, BUFLEN, 0, (struct sockaddr *) &si_other, &slen)) == SOCKET_ERROR)
		{
			printf("SERVER: recvfrom() failed with error code : %d" , WSAGetLastError());
			exit(EXIT_FAILURE);
		}
		
		//print details of the client/peer and the data received
		printf("SERVER: Received packet from %s:%d\n", inet_ntoa(si_other.sin_addr), ntohs(si_other.sin_port));
		printf("SERVER: Data: %s\n" , buf);
		
		// the data string has a dummy header in front of the message text, which we need to remove to find our commands
		ourCmd = &buf[0]; /* this should be the first Character after the header */
		printf("SERVER: Message Payload: %s\n" , ourCmd);
		memset(cmd,'\0', BUFLEN);
		sprintf(cmd, "%s" , ourCmd);
		if (strcmp(cmd, cmd1) == 0) { // SOCKME command received, for this one we do not echo. our response is 'special'
			memset(cmd,'\0', BUFLEN);
			sprintf(cmd, "SOCKET:%d" , xt);
			printf("SERVER: RECEIVED SOCKME command UDP responding: %s and then pausing before send on TCP.\n", cmd);
			recv_len = strlen(cmd);
			if (sendto(su, cmd, recv_len, 0, (struct sockaddr*) &si_other, slen) == SOCKET_ERROR)
			{
				printf("SERVER: sendto() failed with error code : %d" , WSAGetLastError());
				exit(EXIT_FAILURE);
			}		
			// now pause ten seconds and send using our connected TCP client socket to see if teh client gets our data or not.
			Sleep(50);
			printf("SERVER: Sending test value on TCP to see if Client receives it on inherited socket.\n");
			if( send(ct , "SOCKYOU", 7, 0) == SOCKET_ERROR)
			{
				printf("SERVER: TCP Send failed with error code : %d" , WSAGetLastError());
				exit(EXIT_FAILURE);
			}
			puts("SERVER: TCP client socket succesfully sent data, servers part of inheritance test completed (results on client) ");
		}
		else {
			//now reply the client with the same data
			if (sendto(su, buf, recv_len, 0, (struct sockaddr*) &si_other, slen) == SOCKET_ERROR)
			{
				printf("SERVER: sendto() failed with error code : %d" , WSAGetLastError());
				exit(EXIT_FAILURE);
			}
			if (strcmp(cmd, cmd2) == 0) { // EXIT command received AND ECHOED SO EXIT
				puts("SERVER: RECEIVED EXIT command and echoed to client, exiting..");
				Sleep(50);
				closesocket(su);
				closesocket(ct);
				closesocket(xt); // this one may be closed by previous line.. not sure
				closesocket(st);
				WSACleanup();
				return 0;		
			}
		}
	}

	closesocket(su);
	closesocket(ct);
	closesocket(xt); // this one may be closed by previous line.. not sure
	closesocket(st);
	WSACleanup();
	
	return 0;
}
